Package: puremoe
Type: Package
Title: Pubmed Unified REtrieval for Multi-Output Exploration
Version: 1.0.4
Authors@R: c(person("Jason", "Timm", role = c("aut", "cre"), email = "JaTimm@salud.unm.edu"))
Maintainer: Jason Timm <JaTimm@salud.unm.edu>
Description: Access a variety of 'PubMed' data through a single, user-friendly interface, including abstracts, bibliometrics from 'iCite', pubtations from 'PubTator3', and full-text records from 'PMC'.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
LazyDataCompression: xz
Depends: R (>= 3.5)
Imports: rentrez, textshape, xml2, data.table, httr, pbapply, jsonlite,
        rappdirs
Suggests: knitr, rmarkdown, DT, dplyr
VignetteBuilder: knitr
RoxygenNote: 7.3.3
URL: https://github.com/jaytimm/puremoe,
        https://jaytimm.github.io/puremoe/
BugReports: https://github.com/jaytimm/puremoe/issues
NeedsCompilation: no
Packaged: 2026-04-21 16:31:24 UTC; jtimm
Author: Jason Timm [aut, cre]
Repository: CRAN
Date/Publication: 2026-04-21 17:02:08 UTC
Built: R 4.6.0; ; 2026-04-26 09:39:52 UTC; unix

# Generated by roxygen2: do not edit by hand

export(data_mesh_thesaurus)
export(data_mesh_trees)
export(data_pmc_list)
export(endpoint_info)
export(get_records)
export(pmid_to_ftp)
export(pmid_to_pmc)
export(search_pubmed)
import(data.table)
importFrom(data.table,fread)
importFrom(data.table,rbindlist)
importFrom(data.table,setDT)
importFrom(data.table,setorder)
importFrom(httr,GET)
importFrom(httr,content)
importFrom(httr,status_code)
importFrom(jsonlite,stream_in)
importFrom(jsonlite,toJSON)
importFrom(parallel,clusterExport)
importFrom(parallel,detectCores)
importFrom(parallel,makeCluster)
importFrom(parallel,stopCluster)
importFrom(pbapply,pblapply)
importFrom(rappdirs,user_data_dir)
importFrom(rentrez,entrez_fetch)
importFrom(rentrez,entrez_search)
importFrom(rentrez,set_entrez_key)
importFrom(utils,download.file)
importFrom(utils,read.csv)
importFrom(xml2,read_xml)
importFrom(xml2,xml_attr)
importFrom(xml2,xml_find_all)
importFrom(xml2,xml_find_first)
importFrom(xml2,xml_ns_strip)
importFrom(xml2,xml_text)

# puremoe

[![CRAN version](https://www.r-pkg.org/badges/version/puremoe)](https://cran.r-project.org/package=puremoe)
[![CRAN downloads](http://cranlogs.r-pkg.org/badges/last-month/puremoe)](https://cran.r-project.org/package=puremoe)

`puremoe` is an R package that unifies access to PubMed and NLM data behind a single retrieval function. Search PubMed, then pull abstracts, citation metrics, named-entity annotations, full-text articles, or MeSH resources -- all as plain `data.table` objects, all pipe-compatible, no juggling separate APIs or output formats.

---

## Installation

From CRAN:

```r
install.packages("puremoe")
```

Development version:

```r
remotes::install_github("jaytimm/puremoe")
```

---

## The `puremoe` API

### Search

- **`search_pubmed(query, ...)`** -- PubMed query string → character vector of PMIDs. Accepts standard PubMed syntax: field tags (`[TiAb]`, `[MeSH Terms]`, `[DP]`), Boolean operators, wildcards.

### Retrieve

**`get_records(pmids, endpoint, cores, sleep, ncbi_key)`** -- the single retrieval function. Pass PMIDs and name an endpoint; get back a `data.table`.

| endpoint | returns | source |
| --- | --- | --- |
| `pubmed_abstracts` | title, abstract, journal, year, authors, MeSH terms | PubMed E-utilities |
| `pubmed_affiliations` | author × affiliation rows | PubMed E-utilities |
| `icites` | citation count, RCR, NIH percentile, field rate | NIH iCite |
| `pubtations` | gene, disease, chemical, species, mutation annotations | PubTator3 |
| `pmc_fulltext` | full-text sections (requires URLs from `pmid_to_ftp()`) | PMC Open Access |

### ID conversion

- **`pmid_to_pmc(pmids, ...)`** -- PMID → PMC ID + DOI via the NCBI ID Converter.
- **`pmid_to_ftp(pmids, ...)`** -- PMID → PMC ID + open-access FTP URL; pass URLs to `get_records(endpoint = "pmc_fulltext")`.

### MeSH reference data

- **`data_mesh_thesaurus()`** -- MeSH descriptor thesaurus + supplementary concept records; one row per term/synonym.
- **`data_mesh_trees()`** -- MeSH hierarchical tree structure; tree numbers encode the classification path.
- **`data_pmc_list()`** -- PMC open-access file list mapping PMC IDs to file paths and licenses.

### Utilities

- **`endpoint_info(endpoint)`** -- column definitions, rate limits, and notes for each endpoint. Returns a list or JSON; useful for tool schemas in LLM applications.

---

## Vignettes

- [Getting started](https://jaytimm.github.io/puremoe/articles/getting-started.html) -- `search_pubmed()` + all `get_records()` endpoints end-to-end
- [MeSH-guided search](https://jaytimm.github.io/puremoe/articles/mesh-search.html) -- thesaurus lookup, tree navigation, and controlled-vocabulary queries

---

## License

MIT © [Jason Timm](https://github.com/jaytimm)

## Citation

```r
citation("puremoe")
```

## Issues

Report bugs or request features at <https://github.com/jaytimm/puremoe/issues>

#' MeSH Descriptor Frequencies Across PubMed
#'
#' Baseline frequencies for MeSH descriptors computed from a local PostgreSQL
#' mirror of PubMed (April 2026). For each descriptor, counts reflect the
#' number of distinct PMIDs indexed with that term; proportions use the full
#' PubMed corpus of 39,703,112 PMIDs as denominator. Descriptor UI and
#' canonical name are joined from the NLM MeSH thesaurus.
#' Intended as a baseline for MeSH term enrichment analyses against arbitrary
#' PubMed subsets.
#'
#' @format A data.table with 30,521 rows and 4 columns:
#' \describe{
#'   \item{DescriptorUI}{MeSH descriptor unique identifier (e.g.,
#'     \code{D000001})}
#'   \item{DescriptorName}{Canonical MeSH descriptor name}
#'   \item{n_pmids}{Number of distinct PubMed records indexed with this
#'     descriptor}
#'   \item{prop_total}{Proportion of all 39,703,112 PubMed PMIDs indexed
#'     with this descriptor}
#' }
#' @source Computed from \code{mesh_descriptor} table in a local PubMed
#'   PostgreSQL mirror; descriptor metadata from the NLM MeSH Thesaurus
#'   (April 2026).
"data_mesh_frequencies"

#' Extract Named Entities from 'PubMed''s 'PubTator3' Tool
#'
#' This function retrieves named entity annotations from 'PubMed''s 'PubTator3' tool. It fetches data using 'PubMed' IDs and processes the JSON response into a structured format.
#' @param x A vector of 'PubMed' IDs for which annotations are to be retrieved from 'PubTator'.
#' @return A data.table, or NA if no data is available, with columns for 'PubMed' ID, title or abstract location, annotation text, start and end positions of annotations, and annotation types.
#' @importFrom jsonlite stream_in
#' @importFrom data.table rbindlist
#' @noRd
#' 
.get_pubtations <- function(x, sleep) {
  
  con <- tryCatch({
    url(paste0("https://www.ncbi.nlm.nih.gov/research/pubtator3-api/publications/export/biocjson?pmids=", paste(x, collapse = ',')))
  }, error = function(e) {
    message("Unable to connect to PubTator3 API. The resource may be temporarily unavailable.")
    return(NULL)
  })
  
  # If connection failed, return NA
  if (is.null(con)) {
    return(NA)
  }
  
  # Read JSON response safely
  mydata <- tryCatch(jsonlite::stream_in(con, verbose = FALSE), error = function(e) {
    message("Unable to read data from PubTator3 API. The resource may be temporarily unavailable.")
    return(NA)
  })
  
  # Ensure data is valid
  if (!is.data.frame(mydata) || nrow(mydata) == 0) {
    return(NA)  # No data retrieved
  }
  
  # Extract the nested JSON correctly
  if ("PubTator3" %in% names(mydata) && is.list(mydata$PubTator3[[1]])) {
    mydata <- mydata$PubTator3[[1]]
  } else {
    return(NA)  # Unexpected format
  }
  
  # Validate that mydata is a list before proceeding
  if (!is.list(mydata) || !("passages" %in% names(mydata))) {
    return(NA)  # If missing, return NA
  }
  
  # Extract annotations safely
  jj <- list()
  
  for (i in seq_along(mydata$passages)) {
    
    if (!is.list(mydata$passages[[i]])) next  # Skip if not a valid list
    
    pb1 <- mydata$passages[[i]]$annotations
    
    if (!is.list(pb1)) next  # Skip if annotations aren't in a list
    
    names(pb1) <- c("title", "abstract")
    
    # Process title
    if (!is.data.frame(pb1$title) || nrow(pb1$title) == 0) {
      pb1$title <- data.frame(tiab = "title", id = NA, text = NA, locations = NA, identifier = NA, type = NA)
    } else {
      if (!("identifier" %in% names(pb1[["title"]]$infons))) {
        pb1[["title"]]$infons$identifier <- NA
      }
      pb1$title <- cbind(tiab = "title", pb1$title[, c("id", "text", "locations")], identifier = pb1$title$infons$identifier, type = pb1$title$infons$type)
    }
    
    # Process abstract
    if (!is.data.frame(pb1$abstract) || nrow(pb1$abstract) == 0) {
      pb1$abstract <- data.frame(tiab = "abstract", id = NA, text = NA, locations = NA, identifier = NA, type = NA)
    } else {
      if (!("identifier" %in% names(pb1[["abstract"]]$infons))) {
        pb1[["abstract"]]$infons$identifier <- NA
      }
      pb1$abstract <- cbind(tiab = "abstract", pb1$abstract[, c("id", "text", "locations")], identifier = pb1$abstract$infons$identifier, type = pb1$abstract$infons$type)
    }
    
    # Merge title and abstract
    jj[[i]] <- rbind(pb1$title, pb1$abstract)
  }
  
  # If nothing was collected, return NA
  if (length(jj) == 0 || !all(sapply(jj, is.data.frame))) {
    return(NA)
  }
  
  # Convert to data.table
  names(jj) <- mydata$id
  jj0 <- data.table::rbindlist(jj, idcol = "pmid")
  
  # Clean up locations field
  jj0$locations <- as.character(jj0$locations)
  jj0$locations <- gsub("[^[:digit:],]", "", jj0$locations)
  
  # Extract start and end positions
  jj0[, c("start", "length") := data.table::tstrsplit(locations, ",", fixed = TRUE)]
  jj0[, start := as.integer(start)]
  jj0[, end := start + as.integer(length)]
  
  # Remove unnecessary columns
  jj0[, c("length", "locations") := NULL]
  
  Sys.sleep(sleep)
  
  return(jj0)
}

#' Internal: Extract Author Affiliations from 'PubMed' Records
#'
#' Function queries PubMed to extract author affiliations from the fetched records. It processes XML records to obtain detailed information about authors, including their names and affiliations.
#' @param x A character vector with search terms or IDs for fetching records from 'PubMed'.
#' @return A data.table consisting of 'PubMed' IDs, author names, and their affiliations.
#' @importFrom xml2 xml_find_all xml_text
#' @importFrom data.table rbindlist
#' @noRd
#' 
#' 
.get_affiliations <- function (x, sleep) {
  
  # Fetch records from PubMed based on the input x
  records <- .fetch_records(x, sleep)
  
  # Process each PubMed record to extract author affiliations
  z <- lapply(records, function(g){
    
    # Extract the PubMed ID from the record
    pm <- xml2::xml_text(xml2::xml_find_all(g, ".//MedlineCitation/PMID"))
    
    # Find all author elements in the record
    auts <- xml2::xml_find_all(g, ".//Author")
    
    # Process each author element
    cache <- lapply(auts, function(k){
      # Extract and concatenate the last name and first name of the author
      Author <- paste(
        xml2::xml_text(xml2::xml_find_all(k, ".//LastName")),
        xml2::xml_text(xml2::xml_find_all(k, ".//ForeName")),
        sep = ', ')
      
      # Handle cases where the author name is missing
      if(length(Author) == 0){Author <- NA}
      
      # Extract the affiliation information of the author
      Affiliation <- xml2::xml_text(xml2::xml_find_all(k,  ".//Affiliation"))
      
      # Handle cases where the affiliation is missing
      if(length(Affiliation) == 0){Affiliation <- NA}
      
      # Create a data frame with PubMed ID, Author name, and Affiliation
      data.frame(pmid = pm, Author, Affiliation)
    })
    
    # Combine all author information into a single data.table
    data.table::rbindlist(cache)
  })
  
  # Combine all records into one data.table
  x0 <- data.table::rbindlist(z)
  
  # Return the final data.table with author affiliations
  return(x0)
}



# #### clean -- 
# .clean_affiliations <- function(x){
#   
#   x[, Affiliation := sub('^.*?([A-Z])','\\1', Affiliation)]
#   x[, Affiliation := trimws(Affiliation)]
#   x[, Affiliation := gsub('(^.*[[:punct:] ])(.*@.*$)', '\\1', Affiliation)]
#   x[, Affiliation := gsub('(^.*[[:punct:] ])(.*@.*$)', '\\1', Affiliation)]
#   x[, Affiliation := gsub('electronic address.*$|email.*$', '', Affiliation, ignore.case = T)]
#   x[, Affiliation := ifelse(nchar(Affiliation) < 10, NA, Affiliation)]
#   return(x)
# }

#' Convert PubMed IDs (PMIDs) to PMC IDs
#'
#' This function converts a vector of PubMed IDs (PMIDs) to their corresponding
#' PubMed Central (PMC) IDs and DOIs using the NCBI ID Converter API.
#'
#' @param pmids A character or numeric vector of PubMed IDs (PMIDs) to convert.
#' @param batch_size An integer specifying the number of PMIDs to process per
#'   batch. Defaults to 200L. The NCBI API has limitations on batch sizes.
#' @param sleep A numeric value specifying the number of seconds to pause
#'   between API requests. Defaults to 0.5 seconds to respect API rate limits.
#'
#' @return A data.table with columns:
#'   \itemize{
#'     \item \code{pmid}: The input PubMed ID
#'     \item \code{pmcid}: The corresponding PMC ID (NA if not available in PMC)
#'     \item \code{doi}: The corresponding DOI (NA if not available)
#'   }
#'   Results are ordered by PMID. Returns NULL with a message if the API is
#'   unavailable or returns invalid data.
#'
#' @importFrom httr GET status_code
#' @importFrom xml2 read_xml xml_find_all xml_attr
#' @importFrom data.table rbindlist setorder
#'
#' @export
#'
#' @examples
#' \donttest{
#' if (interactive()) {
#'   # Convert a single PMID to PMC ID
#'   result <- pmid_to_pmc("12345678")
#'   
#'   # Convert multiple PMIDs
#'   pmids <- c("12345678", "23456789", "34567890")
#'   result <- pmid_to_pmc(pmids, batch_size = 10, sleep = 1)
#' }
#' }
#'
pmid_to_pmc <- function(pmids, batch_size = 200L, sleep = 0.5) {
  
  # Input validation
  if (missing(pmids) || length(pmids) == 0) {
    message("No PMIDs provided.")
    return(NULL)
  }
  
  if (!(is.character(pmids) || is.numeric(pmids))) {
    message("PMIDs must be character or numeric.")
    return(NULL)
  }
  
  # Convert to character if numeric
  pmids <- as.character(pmids)
  
  # Remove any NA or empty values
  pmids <- pmids[!is.na(pmids) & pmids != ""]
  
  if (length(pmids) == 0) {
    message("No valid PMIDs provided after removing NAs and empty values.")
    return(NULL)
  }
  
  # Ensure batch_size is a positive integer
  batch_size <- as.integer(batch_size[1])
  if (is.na(batch_size) || batch_size < 1) {
    batch_size <- 200L
    message("Invalid batch_size, using default value of 200.")
  }
  
  # Ensure sleep is non-negative
  sleep <- as.numeric(sleep[1])
  if (is.na(sleep) || sleep < 0) {
    sleep <- 0.5
    message("Invalid sleep value, using default value of 0.5.")
  }
  
  # Split PMIDs into batches
  pmid_batches <- split(pmids, ceiling(seq_along(pmids) / batch_size))
  
  # Process each batch
  results <- lapply(pmid_batches, function(batch) {
    # Construct the query URL
    query <- paste0("https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/?",
                    "ids=", paste(batch, collapse = ","),
                    "&idtype=pmid&format=xml")
    
    # Sleep before making request
    Sys.sleep(sleep)
    
    # Make the API request with error handling
    resp <- tryCatch({
      httr::GET(query)
    }, error = function(e) {
      message("Unable to connect to NCBI ID Converter API. The resource may be temporarily unavailable.")
      return(NULL)
    })
    
    # Check if request failed
    if (is.null(resp)) {
      return(NULL)
    }
    
    # Check HTTP status code (fail gracefully instead of using stop_for_status)
    if (httr::status_code(resp) != 200) {
      message("NCBI ID Converter API returned an error. The resource may be temporarily unavailable.")
      return(NULL)
    }
    
    # Parse XML response with error handling
    doc <- tryCatch({
      xml2::read_xml(resp)
    }, error = function(e) {
      message("Unable to parse response from NCBI ID Converter API. The resource may have returned invalid data.")
      return(NULL)
    })
    
    # Check if parsing failed
    if (is.null(doc)) {
      return(NULL)
    }
    
    # Extract records
    recs <- tryCatch({
      xml2::xml_find_all(doc, ".//record")
    }, error = function(e) {
      message("Unable to extract records from API response.")
      return(list())
    })
    
    # If no records found, return NULL for this batch
    if (length(recs) == 0) {
      return(NULL)
    }
    
    # Process each record - just get PMC IDs and DOIs (no URL fetching here)
    dt <- data.table::rbindlist(lapply(recs, function(r) {
      pmid_val <- xml2::xml_attr(r, "pmid")
      pmcid_val <- xml2::xml_attr(r, "pmcid")
      doi_val <- xml2::xml_attr(r, "doi")
      
      list(pmid = pmid_val, pmcid = pmcid_val, doi = doi_val)
    }))
    
    return(dt)
  })
  
  # Filter out NULL results
  results <- results[!sapply(results, is.null)]
  
  # If all batches failed, return NULL
  if (length(results) == 0) {
    message("All API requests failed. Unable to retrieve PMC IDs.")
    return(NULL)
  }
  
  # Combine all results
  result_dt <- data.table::rbindlist(results, fill = TRUE)
  
  # Order by PMID
  data.table::setorder(result_dt, pmid)
  
  return(result_dt[])
}

#' Get 'PubMed' Records
#'
#' Processes XML records obtained from 'PubMed'. It extracts basic bibliographic information and annotations for each record.
#' @param x A character vector with search terms or IDs for fetching records from 'PubMed'.
#' @return A data.table with columns for 'PubMed' IDs, publication year, journal name, article title, abstract, and annotations.
#' @noRd
.get_records <- function (x, sleep) { 
  
  # Fetch records using .fetch_records function and parse XML content
  records <- .fetch_records(x, sleep)
  
  # Process each record to extract basic information and annotations
  parsed_records <- lapply(records, function(x){
    # Extract basic bibliographic information from the record
    basic_info <- .extract_basic(x)
    # Extract annotations (like MeSH terms) from the record
    annotations <- .extract_annotations(x)
    
    # Combine basic information and annotations into a list
    out1 <- list('basic_info' = basic_info, 'annotations' = annotations)
    return(out1)
  })
  
  # Convert the list of basic information into a tidy format
  sum0 <- textshape::tidy_list(x = lapply(parsed_records, '[[', 1),
                               id.name = 'id',
                               content.name = 'varx')
  
  # Reshape the data into a wide format using data.table
  id <- NULL
  sum1 <- data.table::dcast(data = sum0,
                            formula = id ~ attribute,
                            value.var = 'varx')
  
  sum1 <- sum1[order(as.numeric(id))]
  
  # Select and reorder columns for the final output
  sum1 <- sum1[, c('pmid', 
                   'doi',
                   'year', 
                   'pubtype', 
                   'journal', 
                   'articletitle', 
                   'abstract')]
  
  # Add annotations to the data table
  annotations <- NULL
  sum1[, annotations := list(lapply(parsed_records, '[[', 2))] 
  
  # Ensure proper encoding for compatibility
  Encoding(rownames(sum1)) <- 'UTF-8'
  
  # Clean up NA values and return the final data table
  cols <- colnames(sum1)
  sum1[, c(cols) := lapply(.SD, .clean_nas), .SDcols = cols]
  
  return(sum1)
}







#' Extract Basic Information from PubMed Records
#'
#' An internal function that parses XML records from PubMed. It extracts essential bibliographic information such as PubMed ID, journal title, article title, publication year, and abstract.
#' @param g An XML node set representing a single PubMed record.
#' @return A named vector with basic bibliographic information from a PubMed record 
#' @noRd
#' @keywords internal
#' 
.extract_basic <- function(g){

  # Extract the PubMed ID (PMID) from the XML
  pm <- xml2::xml_text(xml2::xml_find_all(g, ".//MedlineCitation/PMID"))

  # Extract DOI
  doi <- xml2::xml_text(xml2::xml_find_all(g, ".//ArticleId[@IdType='doi']"))
  if(length(doi) == 0){
    # Try alternative location for DOI
    doi <- xml2::xml_text(xml2::xml_find_all(g, ".//ELocationID[@EIdType='doi']"))
  }
  if(length(doi) == 0){doi <- NA} else {doi <- doi[1]}  # Use first DOI if multiple

  # Extract the journal title
  a1 <- xml2::xml_text(xml2::xml_find_all(g, ".//Title"))
  a1a <- a1[1]  # In case there are multiple titles, use the first one

  # Extract the article title
  a2 <- xml2::xml_text(xml2::xml_find_all(g, ".//ArticleTitle"))
  
  # Extract publication type
  pub_type <- xml2::xml_text(xml2::xml_find_all(g, ".//PublicationType"))
  # pub_type <- pub_type[1]
  pub_type <- if(length(pub_type) > 1) { paste(pub_type, collapse = " | ") } else 
    if(length(pub_type) == 1) { pub_type } else { NA }

  # Extract the publication year. If 'Year' is not available, use 'MedlineDate' as a fallback
  year <- xml2::xml_text(xml2::xml_find_all(g, ".//PubDate/Year"))
  if(length(year) == 0){
    year <- xml2::xml_text(xml2::xml_find_all(g, ".//PubDate/MedlineDate"))
  }
  # Clean up the year to remove any extra characters or ranges
  year <- gsub(" .+", "", year)
  year <- gsub("-.+", "", year)

  # Extract the abstract text, combining multiple parts if necessary
  abstract <- xml2::xml_text(xml2::xml_find_all(g, ".//Abstract/AbstractText"))
  
  if(length(abstract) > 1){
    abstract <- paste(abstract, collapse = ' ')}
  if(length(abstract) == 0){abstract <- NA}

  abstract <- .reformat_abstract(abstract)
  # Construct the output with the extracted information
  out <- c('pmid' = pm,
           'doi' = doi,
           'journal' = a1a,
           'pubtype' = pub_type,
           'articletitle' = a2,
           'year' = year,
           'abstract' = abstract)

  return(out)
}



#' Extract Annotations from PubMed Records
#'
#' Parses XML records from PubMed to extract annotations such as MeSH terms, chemical names, and keywords. 
#' @param g An XML node set representing a single PubMed record.
#' @return A data frame with annotations extracted from a PubMed record.
#' @noRd
.extract_annotations <- function(g){  
  # Extract the PubMed ID (PMID) from the XML record
  pm <- xml2::xml_text(xml2::xml_find_all(g, ".//MedlineCitation/PMID"))
  
  # Extract MeSH terms (Medical Subject Headings) and their Descriptor UI
  mesh_nodes <- xml2::xml_find_all(g, ".//DescriptorName")
  meshes <- xml2::xml_text(mesh_nodes)
  mesh_ui <- xml2::xml_attr(mesh_nodes, "UI")
  
  # Extract chemical substances names
  chems <- xml2::xml_text(xml2::xml_find_all(g, ".//NameOfSubstance"))
  
  # Extract keywords from the record
  keys <- xml2::xml_text(xml2::xml_find_all(g, ".//Keyword"))
  
  # Combine the extracted data into a single data frame
  df0 <- rbind(
    data.frame(pmid = pm, type = 'MeSH', 
               DescriptorName = if(length(meshes) > 0) {meshes} else {NA}, DescriptorUI = if(length(mesh_ui) > 0) {mesh_ui} else {NA}),
    data.frame(pmid = pm, type = 'Chemistry', 
               DescriptorName = if(length(chems) > 0) {chems} else {NA}, DescriptorUI = NA),
    data.frame(pmid = pm, type = 'Keyword', 
               DescriptorName = if(length(keys) > 0) {keys} else {NA}, DescriptorUI = NA)
  )
  
  # Return the combined annotations data frame
  return(df0)
}



#' Reformat Abstract Text
#'
#' Internal function to reformat an abstract by inserting newlines before each section title.
#' It handles abstracts with or without section titles and trims whitespace from each section.
#' Returns NA if the input is NA.
#'
#' @param abstract A character string representing the abstract text.
#'
#' @return A character string of the reformatted abstract with newlines before each section title, or NA if the input is NA.
#'
#' @noRd
.reformat_abstract <- function(abstract) {
  if (is.na(abstract)) {
    return(NA)
  }
  
  if (!is.character(abstract)) {
    stop("Abstract must be a character string.", call. = FALSE)
  }
  
  # Regular expression to match section titles (e.g., "Methodology and Results:")
  # This pattern matches 1-3 words, each word starting with an uppercase letter or all words being uppercase
  pattern_title <- "(^|\\.\\s+)(([A-Z][a-z]*|[A-Z]+)(\\s([A-Z][a-z]*|[A-Z]+)){0,2}):"
  
  # Use the pattern to insert a newline before each title and split the abstract into sections
  split_abstract <- strsplit(gsub(pattern_title, "\n\\2:", abstract), "\n")[[1]]
  
  # Combine the sections back into a single string
  formatted_abstract <- paste(split_abstract, collapse = "\n")
  
  return(formatted_abstract)
}

#' Search 'PubMed' Records
#'
#' Performs a 'PubMed' search based on a query, optionally filtered by publication years. 
#' Returns a unique set of 'PubMed' IDs matching the query.
#'
#' @param x Character string, the search query.
#' @param start_year Integer, the start year of publication date range (used if `use_pub_years` is TRUE).
#' @param end_year Integer, the end year of publication date range (used if `use_pub_years` is TRUE).
#' @param retmax Integer, maximum number of records to retrieve, defaults to 9999.
#' @param use_pub_years Logical, whether to filter search by publication years, defaults to TRUE.
#' @return Numeric vector of unique PubMed IDs.
#' @importFrom rentrez entrez_search
#' @export
#' @examples
#' ethnob1 <- search_pubmed("ethnobotany", 2010, 2012)
#' 
#' 
search_pubmed <- function(x, 
                          start_year = NULL, 
                          end_year = NULL, 
                          retmax = 9999, 
                          use_pub_years = FALSE) {
  
  if(!is.character(x) || length(x) != 1) {
    stop("x must be a single character string.")
  }
  
  if(use_pub_years) {
    if(is.null(start_year) || is.null(end_year)) {
      stop("start_year and end_year must be provided when use_pub_years is TRUE.")
    }
    if(!is.numeric(start_year) || !is.numeric(end_year) || length(start_year) != 1 || length(end_year) != 1) {
      stop("start_year and end_year must be single integers.")
    }
    if(start_year > end_year) {
      stop("start_year must be less than or equal to end_year.")
    }
    
    all_ids <- vector("list", length = end_year - start_year + 1)
    names(all_ids) <- as.character(start_year:end_year)
    
    for (year in start_year:end_year) {
      query <- paste0(x, " AND ", year, "[Pub Date]")
      all_ids[[as.character(year)]] <- .perform_search(query, retmax)
    }
  } else {
    all_ids <- list(all_years = .perform_search(x, retmax))
  }
  
  return(unique(unlist(all_ids, use.names = FALSE)))
}

#' Internal Function for PubMed Search
#'
#' Handles querying of the PubMed database and returns search results.
#' This function is used internally by 'search_pubmed'.
#'
#' @param query Character string containing the PubMed search query.
#' @param retmax Integer specifying the maximum number of records to retrieve.
#' @noRd
.perform_search <- function(query, retmax) {
  tryCatch({
    result <- rentrez::entrez_search(db = "pubmed", term = query, retmax = retmax, use_history = TRUE)
    if (result$count > 0) {
      return(result$ids)
    } else {
      return(NULL)
    }
  }, error = function(e) {
    warning(sprintf("Failed to retrieve data for query '%s': %s", query, e$message))
    return(NULL)
  })
  Sys.sleep(0.5)
}

#' Fetch Batch of 'PubMed' Records as XML
#'
#' This function attempts to fetch batches of 'PubMed' records in XML format. It retries multiple times in case of failures.
#' @param x A vector of 'PubMed' record identifiers to be fetched.
#' @return A character string with XML content of 'PubMed' records, or an error object in case of failure.
#' @importFrom rentrez entrez_fetch
#' @noRd
#' 
#' 
.fetch_records <- function(x, sleep) {
  # Loop to retry fetching records, with a maximum of 15 attempts
  x1 <- NULL
  for (i in 1:15) {
    # Display the current attempt number
    #message(i)
    
    # Try fetching records using rentrez::entrez_fetch
    x1 <- try({
      rentrez::entrez_fetch(
        db = "pubmed",
        id = x,
        rettype = "xml",
        parsed = FALSE
      )
    }, silent = TRUE)
    
    # Wait before the next attempt
    Sys.sleep(sleep)
    
    # Check if the fetch was successful using inherits(), and if so, break the loop
    if (!inherits(x1, "try-error")) {
      break
    }
  }
  
  # If all attempts failed, fail gracefully with informative message
  if (inherits(x1, "try-error") || is.null(x1)) {
    message("Unable to fetch records from PubMed. The resource may be temporarily unavailable.")
    return(list())
  }
  
  # Return the fetched XML content
  tryCatch({
    doc <- xml2::read_xml(x1)
    xml2::xml_find_all(doc, "//PubmedArticle")
  }, error = function(e) {
    message("Unable to parse records from PubMed. The resource may have returned invalid data.")
    return(list())
  })
}



#' Clean Missing or Invalid Values in Data
#'
#' This function standardizes the representation of missing or invalid values in data by replacing specific character representations of missing data (' ', 'NA', 'n/a', 'n/a.') with R's standard `NA`.
#' @param x A vector that may contain missing or invalid values represented in various formats.
#' @return A vector with standardized missing values represented as `NA`.
#' @noRd
#' 
#' 
.clean_nas <- function(x) {
  
  # Replace specific character representations of missing data with NA
  ifelse(x %in% c(' ', 'NA', 'n/a', 'n/a.') | is.na(x), NA, x) 
}


#' Safe Download Helper Function
#'
#' Downloads a file from a URL with comprehensive error handling.
#' Returns 0 on success, NULL on failure. Never throws errors or warnings.
#'
#' @param url Character string with the URL to download from.
#' @param destfile Character string with the destination file path.
#' @param mode Character string specifying the download mode (default "wb").
#' @return Integer 0 on success, NULL on failure.
#' @noRd
.safe_download <- function(url, destfile, mode = "wb") {
  
  # Suppress all warnings and catch all errors
  result <- tryCatch({
    suppressWarnings({
      utils::download.file(url, destfile, mode = mode, quiet = TRUE)
    })
  }, error = function(e) {
    return(NULL)
  })
  
  # Check if download was successful (returns 0 on success)
  if (is.null(result) || result != 0) {
    return(NULL)
  }
  
  # Verify file was actually created and has content
  if (!file.exists(destfile) || file.size(destfile) == 0) {
    return(NULL)
  }
  
  return(0L)
}

#' Retrieve Data from 'NLM'/'PubMed' databases Based on PMIDs
#'
#' This function retrieves different types of data (like 'PubMed' records, affiliations, 'iCites 'data, etc.) from 'PubMed' based on provided PMIDs. It supports parallel processing for efficiency.
#' 
#' For the 'pmc_fulltext' endpoint, provide full URLs to PMC tar.gz files. 
#' Use \code{\link{pmid_to_pmc}} to convert PMIDs to PMC IDs and full URLs first.
#' 
#' @param pmids A vector of PMIDs for which data is to be retrieved. For 'pmc_fulltext' endpoint, 
#'   provide full URLs instead (e.g., from \code{pmid_to_pmc()$url}).
#' @param endpoint A character vector specifying the type of data to retrieve ('pubtations', 'icites',
#'   'pubmed_affiliations', 'pubmed_abstracts', 'pmc_fulltext').
#' @param cores Number of cores to use for parallel processing (default is 3).
#' @param ncbi_key (Optional) NCBI API key for authenticated access.
#' @param sleep Duration (in seconds) to pause after each batch 
#' @return A data.table containing combined results from the specified endpoint.
#' @importFrom parallel makeCluster stopCluster detectCores clusterExport
#' @importFrom pbapply pblapply
#' @importFrom data.table rbindlist
#' @export
#' @examples
#' pmids <- c("38136652")
#' results <- get_records(pmids, endpoint = "pubmed_abstracts", cores = 1)
#' 
get_records <- function(pmids,
                        endpoint = c('pubtations',
                                     'icites',
                                     'pubmed_affiliations',
                                     'pubmed_abstracts',
                                     'pmc_fulltext'),
                        cores = 3, 
                        sleep = 1,
                        ncbi_key = NULL) {
  
  # Input validation
  if (!(is.character(pmids) || is.numeric(pmids)) || length(pmids) == 0) {
    stop("pmids must be a non-empty vector of characters or numbers")
  }
  
  if (!is.character(endpoint) || length(endpoint) != 1 ||
      !endpoint %in% c('pubtations',
                       'icites',
                       'pubmed_affiliations',
                       'pubmed_abstracts',
                       'pmc_fulltext')) {
    stop("Invalid endpoint. Must be one of 'pubtations', 'icites', 'pubmed_affiliations', 'pubmed_abstracts', 'pmc_fulltext'")
  }
  
  if (!is.numeric(cores)) {
    stop("cores must be numeric")
  }
  
  # Set the NCBI API key for authenticated access if provided
  if (!is.null(ncbi_key)) rentrez::set_entrez_key(ncbi_key)
  
  # Define batch size and the specific task function based on the chosen endpoint
  batch_size <- if (endpoint == "pmc_fulltext") {5} else if (endpoint == "pubtations") {99} else {199}
  task_function <- switch(endpoint,
                          "icites" = .get_icites,
                          "pubtations" = .get_pubtations,
                          "pubmed_affiliations" = .get_affiliations,
                          "pubmed_abstracts" = .get_records,
                          "pmc_fulltext" = .get_pmc,
                          ##
                          stop("Invalid endpoint"))
  
  # Split the PMIDs (or PMC IDs/file paths for pmc_fulltext) into batches for parallel processing
  batches <- split(pmids, ceiling(seq_along(pmids) / batch_size))
  
  if (cores > 1) {
    
    # Parallel processing: Create a cluster and export necessary variables
    clust <- parallel::makeCluster(cores)
    parallel::clusterExport(cl = clust, 
                            varlist = c("task_function", "sleep"), 
                            envir = environment())
    
    # Apply the task function to each batch with the sleep parameter, using parallel processing
    results <- pbapply::pblapply(X = batches, 
                                 FUN = function(batch) task_function(batch, sleep), 
                                 cl = clust)
    parallel::stopCluster(clust)  # Stop the cluster after processing
  } else {
    
    # Sequential processing: Apply the task function to each batch with the sleep parameter
    results <- lapply(batches, function(batch) task_function(batch, sleep))
  }
  
  df_only_list <- results[sapply(results, is.data.frame)]
  # Combine results from all batches into a single data.table
  # Use fill=TRUE to handle cases where batches have different columns
  combined_results <- data.table::rbindlist(df_only_list, fill = TRUE)
  return(combined_results)
}

#' Download and Combine 'MeSH' and Supplemental Thesauruses
#'
#' This function downloads and combines the 'MeSH' (Medical Subject Headings) Thesaurus 
#' and a supplemental concept thesaurus.
#' The data is sourced from specified URLs and stored locally for subsequent use.
#' By default, the data is stored in a temporary directory. Users can opt into 
#' persistent storage by setting `use_persistent_storage` to TRUE and optionally 
#' specifying a path.
#'
#' @param path A character string specifying the directory path where data should 
#' be stored. If not provided and persistent storage is requested, it defaults to 
#' a system-appropriate persistent location managed by `rappdirs`.
#' @param use_persistent_storage A logical value indicating whether to use persistent
#' storage. If TRUE and no path is provided, data will be stored in a system-appropriate 
#' location. Defaults to FALSE, using a temporary directory.
#' @param force_install A logical value indicating whether to force re-downloading 
#' of the data even if it already exists locally.
#' @return A data.table containing the combined MeSH and supplemental thesaurus data.
#' @importFrom rappdirs user_data_dir
#' @importFrom utils download.file
#' @importFrom data.table rbindlist
#' @export
#' @examples
#' \donttest{
#' if (interactive()) {
#'   data <- data_mesh_thesaurus()
#' }
#' }
#' 
data_mesh_thesaurus <- function(path = NULL, 
                                use_persistent_storage = FALSE, 
                                force_install = FALSE) {
  
  # Define the URLs for the MeSH thesaurus and supplemental thesaurus data
  sf <- 'https://github.com/jaytimm/mesh-builds/blob/main/data/data_mesh_thesaurus.rds?raw=true'
  sf2 <- 'https://github.com/jaytimm/mesh-builds/blob/main/data/data_scr_thesaurus.rds?raw=true'
  
  # Determine the directory path based on user preference for persistent storage
  if (use_persistent_storage && is.null(path)) {
    path <- file.path(rappdirs::user_data_dir("puremoe"), "Data")
    if (!dir.exists(path)) {
      dir.create(path, recursive = TRUE)
      message(sprintf("Created persistent directory at: %s", path))
    } else {
      message(sprintf("Directory already exists at: %s", path))
    }
  } else if (is.null(path)) {
    path <- tempdir()
    message("No path provided and persistent storage not requested. Using temporary directory for this session.")
  } else {
    if (!dir.exists(path)) {
      dir.create(path, recursive = TRUE)
      message(sprintf("Created directory at specified path: %s", path))
    } else {
      message(sprintf("Directory already exists at: %s", path))
    }
  }
  
  # Define local file paths for storing the downloaded data
  df <- file.path(path, 'data_mesh_thesaurus.rds')
  df2 <- file.path(path, 'data_scr_thesaurus.rds')
  
  # Check for the existence of the files or force download
  if (!file.exists(df) || force_install) {
    message('Downloading the MeSH thesaurus...')
    download_result <- .safe_download(sf, df, mode = "wb")
    
    # If download failed and no cached file exists
    if (is.null(download_result) && !file.exists(df)) {
      message("Unable to download MeSH thesaurus. The resource may be temporarily unavailable.")
      message("No cached MeSH thesaurus file available. Please check your internet connection and try again.")
    }
  }
  
  if (!file.exists(df2) || force_install) {
    message('Downloading the supplemental concept thesaurus...')
    download_result2 <- .safe_download(sf2, df2, mode = "wb")
    
    # If download failed and no cached file exists
    if (is.null(download_result2) && !file.exists(df2)) {
      message("Unable to download supplemental concept thesaurus. The resource may be temporarily unavailable.")
      message("No cached supplemental thesaurus file available. Please check your internet connection and try again.")
    }
  }
  
  # If either file doesn't exist, return NULL
  if (!file.exists(df) || !file.exists(df2)) {
    message("One or both thesaurus files are not available.")
    return(NULL)
  }
  
  # Read the downloaded RDS files
  a1 <- readRDS(df)
  a2 <- readRDS(df2)
  
  # Ensure the column names are consistent between the two data sets
  colnames(a2) <- colnames(a1)
  
  # Combine the data using data.table's rbindlist
  combined_data <- rbindlist(list(a1, a2))
  
  return(combined_data)
}

#' Get Information About Available Endpoints
#'
#' This function provides detailed information about the available endpoints
#' in the package, including column descriptions, parameters, rate limits, and usage notes.
#'
#' @param endpoint Character string specifying which endpoint to get information about.
#'   If NULL (default), returns a list of all available endpoints.
#' @param format Character string specifying the output format. Either "list" (default)
#'   or "json" for JSON-formatted output.
#'
#' @return If \code{endpoint} is NULL, returns a character vector of available endpoint names.
#'   If \code{endpoint} is specified, returns a list (or JSON string) with detailed information
#'   about that endpoint including description, columns, parameters, rate limits, and notes.
#'
#' @importFrom jsonlite toJSON
#'
#' @export
#'
#' @examples
#' \donttest{
#' if (interactive()) {
#'   # List all available endpoints
#'   endpoint_info()
#'   
#'   # Get information about a specific endpoint
#'   endpoint_info("pubmed_abstracts")
#'   
#'   # Get information in JSON format
#'   endpoint_info("icites", format = "json")
#' }
#' }
#'
endpoint_info <- function(endpoint = NULL, format = c("list", "json")) {
  format <- match.arg(format)
  
  schemas <- list(
    pubmed_abstracts = list(
      description = "PubMed article metadata and abstracts",
      returns = "data.frame",
      columns = list(
        pmid = "PubMed ID (character)",
        doi = "Digital Object Identifier (character)",
        year = "Publication year (integer)",
        pubtype = "Publication type(s) (character)",
        journal = "Journal name (character)",
        articletitle = "Article title (character)",
        abstract = "Abstract text (character)",
        annotations = "MeSH terms and keywords (list-column)"
      ),
      parameters = list(cores = "parallel workers", sleep = "delay between requests (seconds)"),
      rate_limit = "NCBI E-utilities: 3/sec without key, 10/sec with key",
      notes = "Use ncbi_key for higher rate limits. Primary source of article metadata."
    ),
    
    pubmed_affiliations = list(
      description = "Author affiliations from PubMed",
      returns = "data.frame",
      columns = list(
        pmid = "PubMed ID (character)",
        Author = "Author name (character)",
        Affiliation = "Institutional affiliation (character)"
      ),
      parameters = list(cores = "parallel workers", sleep = "delay between requests"),
      rate_limit = "NCBI E-utilities: 3/sec without key, 10/sec with key",
      notes = "One row per author; multiple affiliations possible"
    ),
    
    icites = list(
      description = "NIH iCite citation metrics and influence scores",
      returns = "data.frame",
      columns = list(
        pmid = "PubMed ID - join key to link with pubmed_abstracts (character)",
        citation_count = "Total citations received (integer)",
        relative_citation_ratio = "RCR: field-adjusted citation rate comparing to NIH baseline (numeric)",
        nih_percentile = "Percentile rank vs NIH-funded publications (numeric)",
        field_citation_rate = "Expected citation rate for article's co-citation field (numeric)",
        is_research_article = "Flag for primary research articles (logical)",
        is_clinical = "Flag for clinical articles (logical)",
        provisional = "Flag indicating RCR is provisional due to recent publication (logical)",
        citation_net = "Citation network edge list: 'from' and 'to' PMIDs within result set (list-column)",
        cited_by_clin = "PMIDs of clinical articles citing this paper (character/list)"
      ),
      parameters = list(cores = "parallel workers", sleep = "delay between requests"),
      rate_limit = "Relatively permissive",
      notes = "Join to pubmed_abstracts on pmid for complete metadata (title, journal, authors, etc. not included to avoid redundancy). citation_net enables intra-corpus network analysis."
    ),
    
    pubtations = list(
      description = "PubTator entity annotations (genes, diseases, chemicals, etc.)",
      returns = "data.frame",
      columns = list(
        pmid = "PubMed ID (character)",
        tiab = "Title/abstract text (character)",
        id = "Annotation ID (character)",
        text = "Annotated text span (character)",
        identifier = "Database identifier (character)",
        type = "Entity type: Gene, Disease, Chemical, Species, Mutation (character)",
        start = "Start position in text (integer)",
        end = "End position in text (integer)"
      ),
      parameters = list(cores = "parallel workers"),
      rate_limit = "Moderate",
      notes = "One row per annotation; multiple annotations per article. Provides named entity recognition output."
    ),
    
    pmc_fulltext = list(
      description = "Full-text articles from PubMed Central",
      returns = "data.frame",
      columns = list(
        pmid = "PubMed ID (character)",
        section = "Section heading (character)",
        text = "Section text content (character)"
      ),
      parameters = list(cores = "parallel workers"),
      input = "Requires FTP URLs from pmid_to_ftp()",
      rate_limit = "NCBI FTP: be respectful",
      notes = "One row per section; use after pmid_to_ftp() to get URLs. Not all PMIDs have PMC full text available."
    )
  )
  
  # Return available endpoints
  if (is.null(endpoint)) {
    result <- names(schemas)
    if (format == "json") {
      return(jsonlite::toJSON(list(available_endpoints = result), 
                              pretty = TRUE, auto_unbox = TRUE))
    }
    return(result)
  }
  
  # Validate endpoint
  if (!endpoint %in% names(schemas)) {
    stop("Unknown endpoint. Available: ", paste(names(schemas), collapse = ", "))
  }
  
  result <- schemas[[endpoint]]
  
  # Return in requested format
  if (format == "json") {
    return(jsonlite::toJSON(result, pretty = TRUE, auto_unbox = TRUE))
  }
  
  result
}

#' Download and Load 'MeSH' Trees Data
#'
#' This function downloads and loads the 'MeSH' (Medical Subject Headings) Trees data.
#' 
#' The data is sourced from specified URLs and stored locally for subsequent use.
#' By default, the data is stored in a temporary directory. Users can opt into 
#' persistent storage by setting `use_persistent_storage` to TRUE and optionally 
#' specifying a path.
#'
#' @param path A character string specifying the directory path where data should 
#' be stored. If not provided and persistent storage is requested, it defaults to 
#' a system-appropriate persistent location managed by `rappdirs`.
#' @param use_persistent_storage A logical value indicating whether to use persistent
#' storage. If TRUE and no path is provided, data will be stored in a system-appropriate 
#' location. Defaults to FALSE, using a temporary directory.
#' @param force_install A logical value indicating whether to force re-downloading 
#' of the data even if it already exists locally.
#' @return A data frame containing the MeSH Trees data.
#' @importFrom rappdirs user_data_dir
#' @importFrom utils download.file
#' @export
#' @examples
#' \donttest{
#' if (interactive()) {
#'   data <- data_mesh_trees()
#' }
#' }
#' 
data_mesh_trees <- function(path = NULL, 
                            use_persistent_storage = FALSE, 
                            force_install = FALSE) {
  
  # Define the URL for the MeSH trees data
  sf <- 'https://github.com/jaytimm/mesh-builds/blob/main/data/data_mesh_trees.rds?raw=true'
  
  # Check if user opts for persistent storage and no path is provided
  if (use_persistent_storage && is.null(path)) {
    path <- file.path(rappdirs::user_data_dir("puremoe"), "Data")
    if (!dir.exists(path)) {
      dir.create(path, recursive = TRUE)
      message(sprintf("Created persistent directory at: %s", path))
    }
  } else if (is.null(path)) {
    # Default to tempdir() for temporary storage
    path <- tempdir()
    message("No path provided and persistent storage not requested. Using temporary directory for this session.")
  } else {
    # Check if the specified path exists; if not, create it
    if (!dir.exists(path)) {
      dir.create(path, recursive = TRUE)
      message(sprintf("Created directory at specified path: %s", path))
    }
  }
  
  # Determine the local file path for storing the data
  df <- file.path(path, 'data_mesh_trees.rds')
  
  # Check if the file exists or if forced download is requested
  if (!file.exists(df) || force_install) {
    # Download the MeSH trees data
    message('Downloading MeSH trees...')
    download_result <- .safe_download(sf, df, mode = "wb")
    
    # If download failed and no cached file exists, return NULL
    if (is.null(download_result) && !file.exists(df)) {
      message("Unable to download MeSH trees data. The resource may be temporarily unavailable.")
      message("No cached file available. Please check your internet connection and try again.")
      return(NULL)
    }
  }
  
  # If file doesn't exist at this point, return NULL
  if (!file.exists(df)) {
    message("MeSH trees data is not available.")
    return(NULL)
  }
  
  # Read and return the downloaded RDS file
  a1 <- readRDS(df)
  return(a1)
}

#' Internal: Fetch Data from 'iCite' Database
#'
#' This internal function is designed to scrape data from the 'iCite' database, a bibliometric tool provided by the 'NIH'. It constructs a URL to query 'iCite' with specified 'PubMed' IDs and retrieves citation metrics and other related data.
#' @param x A vector of 'PubMed' IDs for which data is to be fetched from the 'iCite' database.
#' @return A data.frame consisting of the data retrieved from iCite, formatted as CSV.
#' @importFrom httr GET content
#' @importFrom utils read.csv
#' @noRd
#' 

.fetch_icites <- function(x, sleep){
  
  # Construct the URL for the iCite API call, including the PubMed IDs (x)
  url0 <- tryCatch({
    httr::GET(paste0("https://icite.od.nih.gov/api/pubs?pmids=",
                     paste(x, collapse = ","),
                     "&format=csv"))
  }, error = function(e) {
    message("Unable to connect to iCite API. The resource may be temporarily unavailable.")
    return(NULL)
  })
  
  # If GET request failed, return empty data.frame
  if (is.null(url0)) {
    return(data.frame(pmid = x))
  }
  
  # Check HTTP status code
  if (httr::status_code(url0) != 200) {
    message("iCite API returned an error. The resource may be temporarily unavailable.")
    return(data.frame(pmid = x))
  }
  
  # Read the content of the response as a CSV.
  csv_ <- tryCatch({
    utils::read.csv(textConnection(
      httr::content(url0,
                    "text",
                    encoding = "UTF-8")),
      encoding = "UTF-8")
  }, error = function(e) {
    message("Unable to parse data from iCite API. The resource may have returned invalid data.")
    return(data.frame(pmid = x))
  })
  
  Sys.sleep(sleep)
  # Return the CSV content as a data.frame
  return(csv_)
}




#' Process and Structure Data from iCite
#'
#' Function processes and structures the data obtained via `.fetch_icites`.
#' @param x A vector of PubMed IDs for which data has been fetched from the iCite database.
#' @return A data.table enhanced with citation network information and cleaned reference and citation data.
#' @importFrom data.table setDT
#' @noRd
#' 
#' 
# .get_icites <- function(x, sleep){
#   
#   # Fetch data from iCite using the PubMed IDs provided
#   pmiddf <- .fetch_icites(x, sleep)
#   
#   # Extract the PubMed IDs for reference
#   gots <- pmiddf$pmid
#   
#   # Convert pmiddf to a data.table for efficient data manipulation
#   data.table::setDT(pmiddf)
#   
#   # Clean and format the 'ref_count' column
#   ref_count <- NULL
#   pmiddf[, ref_count := ifelse(is.null(references)|is.na(references), NULL, references)]
#   
#   # Process 'references' and 'cited_by' columns, handling empty or NA values
#   pmiddf[, references := ifelse(nchar(references) == 0|is.na(references), '99', references)]
#   pmiddf[, cited_by := ifelse(nchar(cited_by) == 0|is.na(cited_by), '99', cited_by)]
#   
#   # Split the 'cited_by' and 'references' columns into lists
#   cited_by <- strsplit(pmiddf$cited_by, split = " ")
#   references <- strsplit(pmiddf$references, split = " ")
#   rs <- strsplit(pmiddf$ref_count, split = " ")
#   
#   # Build a data frame for references
#   doc_id <- NULL
#   from <- NULL
#   refs <- data.table::data.table(doc_id = rep(gots, sapply(references, length)),
#                                  from = rep(gots, sapply(references, length)),
#                                  to = unlist(references))
#   # Replace placeholder '99' with NA
#   refs[refs == 99] <- NA
#   
#   # Aggregate reference data and convert to a data.table
#   refs0 <- refs[, list(references = .N), by = list(from)]
#   
#   # Build a data frame for cited_by data
#   cited <- data.frame(doc_id = rep(gots, sapply(cited_by, length)),
#                       from = unlist(cited_by),
#                       to = rep(gots, sapply(cited_by, length)))
#   # Replace placeholder '99' with NA
#   cited[cited == 99] <- NA
#   
#   # Combine references and cited_by data
#   f1 <- rbind(refs, cited)
#   # Aggregate the combined data and format as a list within a data.table
#   f2 <- data.table::setDT(f1)[, list(references = list(.SD)), by = doc_id]
#   
#   # Add citation network data to pmiddf
#   citation_net <- NULL
#   pmiddf[, citation_net := f2$references]
#   # Calculate and add reference count
#   pmiddf[, ref_count := sapply(rs, length)]
#   # Remove the original 'cited_by' and 'references' columns
#   pmiddf[, c('cited_by', 'references') := NULL]
#   
#   # Return the processed data table
#   pmiddf[, c(1, 6:25)]
# }

.get_icites <- function(x, sleep = 0.25) {
  
  pmiddf <- .fetch_icites(x, sleep)
  
  # If fetch failed, return NULL
  if (is.null(pmiddf) || nrow(pmiddf) == 0) {
    return(NULL)
  }
  
  data.table::setDT(pmiddf)
  
  # Normalize pmid column - use X_id if pmid is missing or NA
  if ("X_id" %in% names(pmiddf)) {
    if (!"pmid" %in% names(pmiddf)) {
      # If pmid doesn't exist, create it from X_id
      pmiddf[, pmid := X_id]
    } else {
      # If pmid exists but has NA values, fill from X_id
      pmiddf[is.na(pmid) | pmid == "", pmid := X_id]
    }
  }
  
  # Ensure pmid column exists
  if (!"pmid" %in% names(pmiddf)) {
    message("Warning: No pmid or X_id column found in iCite response.")
    return(NULL)
  }
  
  # normalize field names to old expectations
  if ("citedPmids" %in% names(pmiddf) && !"references" %in% names(pmiddf)) {
    data.table::setnames(pmiddf, "citedPmids", "references")
  }
  if ("citedByPmids" %in% names(pmiddf) && !"cited_by" %in% names(pmiddf)) {
    data.table::setnames(pmiddf, "citedByPmids", "cited_by")
  }
  
  # Ensure required columns exist for citation network processing
  has_references <- "references" %in% names(pmiddf)
  has_cited_by <- "cited_by" %in% names(pmiddf)
  
  if (!has_references) {
    pmiddf[, references := NA_character_]
  }
  if (!has_cited_by) {
    pmiddf[, cited_by := NA_character_]
  }
  
  # Extract PMIDs for citation network processing
  gots <- pmiddf$pmid
  gots <- as.character(gots)  # Ensure character type
  
  # Process citation networks if we have the data
  if (has_references || has_cited_by) {
    pmiddf[, ref_count := ifelse(is.null(references)|is.na(references), NULL, references)]
    pmiddf[, references := ifelse(nchar(references) == 0|is.na(references), '99', references)]
    pmiddf[, cited_by := ifelse(nchar(cited_by) == 0|is.na(cited_by), '99', cited_by)]
    
    cited_by  <- strsplit(pmiddf$cited_by,  split = " ")
    references <- strsplit(pmiddf$references, split = " ")
    rs <- strsplit(pmiddf$ref_count, split = " ")
    
    refs <- data.table::data.table(
      doc_id = rep(gots, sapply(references, length)),
      from   = rep(gots, sapply(references, length)),
      to     = unlist(references)
    )
    refs[refs == 99] <- NA
    refs0 <- refs[, .(references = .N), by = .(from)]
    
    cited <- data.table::data.table(
      doc_id = rep(gots, sapply(cited_by, length)),
      from   = unlist(cited_by),
      to     = rep(gots, sapply(cited_by, length))
    )
    cited[cited == 99] <- NA
    
    f1 <- rbind(refs, cited)
    if (nrow(f1) > 0) {
      f2 <- data.table::setDT(f1)[, .(references = list(.SD)), by = doc_id]
      pmiddf[, citation_net := f2$references[match(gots, f2$doc_id)]]
      pmiddf[, ref_count := sapply(rs, length)]
    } else {
      pmiddf[, citation_net := list(list())]
      pmiddf[, ref_count := 0L]
    }
    
    # Remove temporary columns if they existed before
    if (has_references || has_cited_by) {
      cols_to_remove <- c("references", "cited_by")
      cols_to_remove <- cols_to_remove[cols_to_remove %in% names(pmiddf)]
      if (length(cols_to_remove) > 0) {
        pmiddf[, (cols_to_remove) := NULL]
      }
    }
  } else {
    # No citation network data, just set defaults
    pmiddf[, citation_net := list(list())]
    pmiddf[, ref_count := 0L]
  }
  
  # Define columns to keep (only select those that exist)
  icites_keep <- c(
    "pmid",
    "citation_count",
    "relative_citation_ratio",
    "nih_percentile",
    "field_citation_rate",
    "is_research_article",
    "is_clinical",
    "provisional",
    "citation_net",
    "cited_by_clin"
  )
  
  # Select only columns that exist in the data.table
  cols_to_keep <- icites_keep[icites_keep %in% names(pmiddf)]
  
  # If pmid is not in the keep list but exists, ensure it's included
  if (!"pmid" %in% cols_to_keep && "pmid" %in% names(pmiddf)) {
    cols_to_keep <- c("pmid", cols_to_keep)
  }
  
  # Select only the columns that exist
  if (length(cols_to_keep) > 0) {
    pmiddf <- pmiddf[, .SD, .SDcols = cols_to_keep]
  }
  
  return(pmiddf[])
}

#' Scrape Full Text Entries from 'PubMed Central' (PMC)
#'
#' Retrieves full-text articles from the PMC Cloud Service on AWS S3
#' using direct XML URLs (e.g., from \code{\link{pmid_to_ftp}}).
#' @param x A vector of HTTPS URLs to PMC XML files.
#' @return A data.table with columns for pmid, section, and text.
#' @importFrom xml2 read_xml xml_find_all xml_find_first xml_text xml_ns_strip
#' @noRd
#'
#'
.get_pmc <- function(x, sleep) {

  flist <- list()

  for (q in seq_along(x)) {

    fn <- x[q]

    x0 <- tryCatch({
      xml2::read_xml(fn)
    }, error = function(e) {
      NULL
    })

    if (!is.null(x0)) {
      pmid <- xml2::xml_text(
        xml2::xml_find_first(
          x0,
          ".//article-meta//article-id[@pub-id-type='pmid']"
        )
      )

      body <- xml2::xml_find_first(x0, ".//body")

      if (!is.null(body) && length(body) > 0) {
        secs <- xml2::xml_find_all(body, "./sec")

        if (length(secs) > 0) {
          header_titles <- lapply(secs, function(sec) {
            title_node <- xml2::xml_find_first(sec, "./title")
            if (length(title_node) > 0) xml2::xml_text(title_node)
            else NA_character_
          })

          text <- lapply(secs, xml2::xml_text)

          df <- data.frame(pmid,
                           section = unlist(header_titles),
                           text = unlist(text),
                           row.names = NULL)

          df$text <- gsub("([a-z]+)([A-Z])", "\\1\n\\2", df$text)

          flist[[q]] <- df
        }
      }
    }
    Sys.sleep(sleep)
  }

  data.table::rbindlist(flist)
}

#' Download and Load 'PMC' Open Access File List
#'
#' This function downloads and loads the 'PMC' (PubMed Central) Open Access file list.
#' The file list contains mappings between PMC IDs, PMIDs, and file paths for 
#' open access articles available for download.
#' 
#' The data is sourced from NCBI's FTP server and stored locally for subsequent use.
#' By default, the data is stored in a temporary directory. Users can opt into 
#' persistent storage by setting `use_persistent_storage` to TRUE and optionally 
#' specifying a path.
#'
#' @param path A character string specifying the directory path where data should 
#' be stored. If not provided and persistent storage is requested, it defaults to 
#' a system-appropriate persistent location managed by `rappdirs`.
#' @param use_persistent_storage A logical value indicating whether to use persistent
#' storage. If TRUE and no path is provided, data will be stored in a system-appropriate 
#' location. Defaults to FALSE, using a temporary directory.
#' @param force_install A logical value indicating whether to force re-downloading 
#' of the data even if it already exists locally.
#' @return A data.table containing the PMC file list with columns: file_path, 
#' citation, pmcid, pmid, and license_code.
#' @importFrom rappdirs user_data_dir
#' @importFrom utils download.file
#' @importFrom data.table fread
#' @export
#' @examples
#' \donttest{
#' if (interactive()) {
#'   data <- data_pmc_list()
#' }
#' }
#' 
data_pmc_list <- function(path = NULL, 
                         use_persistent_storage = FALSE, 
                         force_install = FALSE) {
  
  # Define the URL for the PMC file list
  sf <- 'https://ftp.ncbi.nlm.nih.gov/pub/pmc/oa_file_list.txt'
  
  # Determine the directory path based on user preference for persistent storage
  if (use_persistent_storage && is.null(path)) {
    path <- file.path(rappdirs::user_data_dir("puremoe"), "Data")
    if (!dir.exists(path)) {
      dir.create(path, recursive = TRUE)
      message(sprintf("Created persistent directory at: %s", path))
    } else {
      message(sprintf("Directory already exists at: %s", path))
    }
  } else if (is.null(path)) {
    path <- tempdir()
    message("No path provided and persistent storage not requested. Using temporary directory for this session.")
  } else {
    if (!dir.exists(path)) {
      dir.create(path, recursive = TRUE)
      message(sprintf("Created directory at specified path: %s", path))
    } else {
      message(sprintf("Directory already exists at: %s", path))
    }
  }
  
  # Define local file path for storing the downloaded data
  df <- file.path(path, 'oa_file_list.txt')
  
  # Check for the existence of the file or force download
  if (!file.exists(df) || force_install) {
    message('Downloading the PMC Open Access file list...')
    download_result <- .safe_download(sf, df, mode = "wb")
    
    # If download failed and no cached file exists
    if (is.null(download_result) && !file.exists(df)) {
      message("Unable to download PMC file list. The resource may be temporarily unavailable.")
      message("No cached PMC file list available. Please check your internet connection and try again.")
      return(NULL)
    }
  }
  
  # If file doesn't exist, return NULL
  if (!file.exists(df)) {
    message("PMC file list is not available.")
    return(NULL)
  }
  
  # Read the file as a tab-separated file
  # The file format is: file_path \t citation \t pmcid \t pmid \t license_code
  dt <- tryCatch({
    data.table::fread(df, sep = "\t", header = FALSE, 
                      col.names = c("file_path", "citation", "pmcid", "pmid", "license_code"),
                      na.strings = c("", "NA"))
  }, error = function(e) {
    message("Unable to parse PMC file list. The file may be corrupted.")
    return(NULL)
  })
  
  # Clean up PMID column (remove "PMID:" prefix if present)
  if (!is.null(dt) && "pmid" %in% names(dt)) {
    dt[, pmid := gsub("^PMID:", "", pmid)]
  }
  
  # Construct full URLs from file paths
  if (!is.null(dt) && "file_path" %in% names(dt)) {
    dt[, url := paste0("https://ftp.ncbi.nlm.nih.gov/pub/pmc/", file_path)]
  }
  
  return(dt)
}

#' Convert PubMed IDs (PMIDs) to PMC IDs and Full-Text URLs
#'
#' This function converts PMIDs to PMC IDs, then fetches the full-text file URLs
#' from the PMC Open Access service. It combines both steps into a single workflow.
#'
#' @param pmids A character or numeric vector of PubMed IDs (PMIDs) to convert.
#' @param batch_size An integer specifying the number of PMIDs to process per
#'   batch for ID conversion. Defaults to 200L. The NCBI API has limitations on batch sizes.
#' @param sleep A numeric value specifying the number of seconds to pause
#'   between API requests for ID conversion (Step 1). Defaults to 0.5 seconds.
#'   For OA API calls (Step 2), sleep time is automatically adjusted based on
#'   rate limits: 0.11s with API key (10 req/sec), 0.34s without (3 req/sec).
#' @param verbose Logical, whether to print progress messages. Defaults to FALSE.
#' @param ncbi_key (Optional) NCBI API key for authenticated access.
#'
#' @return A data.table with columns:
#'   \itemize{
#'     \item \code{pmid}: The input PubMed ID
#'     \item \code{pmcid}: The corresponding PMC ID
#'     \item \code{doi}: The corresponding DOI (NA if not available)
#'     \item \code{url}: The full HTTPS URL for downloading PMC full text
#'   }
#'   Results are filtered to only include rows with valid URLs (open access articles),
#'   ordered by PMID. Returns NULL with a message if the API is unavailable or returns invalid data.
#'
#' @importFrom httr GET status_code
#' @importFrom xml2 read_xml xml_find_all xml_attr xml_find_first
#' @importFrom data.table rbindlist setorder
#' @importFrom rentrez set_entrez_key
#'
#' @export
#'
#' @examples
#' \donttest{
#' if (interactive()) {
#'   # Convert PMIDs to PMC IDs and get full-text URLs
#'   result <- pmid_to_ftp(c("11250746", "11573492"))
#' }
#' }
#'
pmid_to_ftp <- function(pmids, batch_size = 200L, sleep = 0.5, verbose = FALSE, ncbi_key = NULL) {
  
  # Input validation
  if (missing(pmids) || length(pmids) == 0) {
    message("No PMIDs provided.")
    return(NULL)
  }
  
  if (!(is.character(pmids) || is.numeric(pmids))) {
    message("PMIDs must be character or numeric.")
    return(NULL)
  }
  
  # Set the NCBI API key for authenticated access if provided
  if (!is.null(ncbi_key)) rentrez::set_entrez_key(ncbi_key)
  
  # Convert to character if numeric
  pmids <- as.character(pmids)
  
  # Remove any NA or empty values
  pmids <- pmids[!is.na(pmids) & pmids != ""]
  
  if (length(pmids) == 0) {
    message("No valid PMIDs provided after removing NAs and empty values.")
    return(NULL)
  }
  
  # Ensure batch_size is a positive integer
  batch_size <- as.integer(batch_size[1])
  if (is.na(batch_size) || batch_size < 1) {
    batch_size <- 200L
    message("Invalid batch_size, using default value of 200.")
  }
  
  # Ensure sleep is non-negative
  sleep <- as.numeric(sleep[1])
  if (is.na(sleep) || sleep < 0) {
    sleep <- 0.5
    message("Invalid sleep value, using default value of 0.5.")
  }
  
  # Step 1: Convert PMIDs to PMC IDs (batched)
  if (verbose) message("Step 1: Converting PMIDs to PMC IDs...")
  
  pmid_batches <- split(pmids, ceiling(seq_along(pmids) / batch_size))
  
  id_results <- lapply(pmid_batches, function(batch) {
    # Construct the query URL
    query <- paste0("https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/?",
                    "ids=", paste(batch, collapse = ","),
                    "&idtype=pmid&format=xml")
    
    Sys.sleep(sleep)
    
    # Make the API request with error handling
    resp <- tryCatch({
      httr::GET(query)
    }, error = function(e) {
      if (verbose) message("Unable to connect to NCBI ID Converter API.")
      return(NULL)
    })
    
    if (is.null(resp) || httr::status_code(resp) != 200) {
      return(NULL)
    }
    
    # Parse XML response
    doc <- tryCatch({
      xml2::read_xml(resp)
    }, error = function(e) {
      return(NULL)
    })
    
    if (is.null(doc)) {
      return(NULL)
    }
    
    # Extract records
    recs <- tryCatch({
      xml2::xml_find_all(doc, ".//record")
    }, error = function(e) {
      return(list())
    })
    
    if (length(recs) == 0) {
      return(NULL)
    }
    
    # Process each record
    dt <- data.table::rbindlist(lapply(recs, function(r) {
      pmid_val <- xml2::xml_attr(r, "pmid")
      pmcid_val <- xml2::xml_attr(r, "pmcid")
      doi_val <- xml2::xml_attr(r, "doi")
      
      list(pmid = pmid_val, pmcid = pmcid_val, doi = doi_val)
    }))
    
    return(dt)
  })
  
  # Filter out NULL results
  id_results <- id_results[!sapply(id_results, is.null)]
  
  if (length(id_results) == 0) {
    message("All ID conversion requests failed. Unable to retrieve PMC IDs.")
    return(NULL)
  }
  
  # Combine ID conversion results
  id_dt <- data.table::rbindlist(id_results, fill = TRUE)
  
  # Step 2: Get S3 XML URLs for PMC IDs using the PMC Cloud Service on AWS
  if (verbose) message("Step 2: Fetching full-text URLs from PMC S3 Cloud Service...")

  pmc_ids_with_urls <- id_dt[!is.na(pmcid) & pmcid != "", ]

  if (nrow(pmc_ids_with_urls) > 0) {
    oa_sleep <- if (!is.null(ncbi_key)) 0.11 else 0.34

    url_results <- lapply(seq_len(nrow(pmc_ids_with_urls)), function(i) {
      pmcid_val <- pmc_ids_with_urls$pmcid[i]

      # Query S3 list API to find available article versions
      s3_list_url <- paste0("https://pmc-oa-opendata.s3.amazonaws.com/?prefix=",
                            pmcid_val, ".&delimiter=/")

      Sys.sleep(oa_sleep)

      s3_resp <- tryCatch({
        httr::GET(s3_list_url)
      }, error = function(e) {
        return(NULL)
      })

      full_url <- NA_character_

      if (!is.null(s3_resp) && httr::status_code(s3_resp) == 200) {
        s3_doc <- tryCatch({
          xml2::read_xml(s3_resp)
        }, error = function(e) {
          return(NULL)
        })

        if (!is.null(s3_doc)) {
          xml2::xml_ns_strip(s3_doc)
          prefixes <- xml2::xml_text(xml2::xml_find_all(s3_doc, ".//CommonPrefixes/Prefix"))

          if (length(prefixes) > 0) {
            # Take the last prefix (latest version), strip trailing slash
            latest_prefix <- sub("/$", "", prefixes[length(prefixes)])
            full_url <- paste0("https://pmc-oa-opendata.s3.amazonaws.com/",
                               latest_prefix, "/", latest_prefix, ".xml")
          }
        }
      }

      list(pmcid = pmcid_val, url = full_url)
    })
    
    url_dt <- data.table::rbindlist(url_results)
    
    # Add url column to id_dt if it doesn't exist
    if (!"url" %in% names(id_dt)) {
      id_dt[, url := NA_character_]
    }
    
    # Merge URLs back into main data.table using data.table syntax
    id_dt[url_dt, url := i.url, on = "pmcid"]
  } else {
    # No PMC IDs found, add empty url column
    id_dt[, url := NA_character_]
  }
  
  # Order by PMID
  data.table::setorder(id_dt, pmid)
  
  # Filter out rows where url is NA (no open access available)
  id_dt <- id_dt[!is.na(url)]
  
  return(id_dt[])
}

---
title: "MeSH Tables"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{MeSH Tables}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  comment = "#>"
)
```

`puremoe` ships three MeSH reference tables: a thesaurus of descriptors and entry terms, a tree of hierarchical classifications, and a frequency table of descriptor occurrence across PubMed.

```{r libs}
library(puremoe)
library(dplyr)
library(DT)
```

## MeSH thesaurus

`data_mesh_thesaurus()` downloads and combines the MeSH Descriptor Thesaurus and Supplementary Concept Records (SCR). One row per term, including synonyms and entry terms for each descriptor.

```{r thesaurus}
thesaurus <- puremoe::data_mesh_thesaurus()
```

```{r thesaurus-table}
thesaurus |>
  head(20) |>
  DT::datatable(rownames = FALSE, options = list(scrollX = TRUE))
```

## MeSH trees

`data_mesh_trees()` provides the hierarchical classification structure. Each descriptor can appear in multiple branches; `tree_location` encodes the full path (e.g., `I01.880.604` = Social Sciences > Political Science > Political Systems).

```{r trees}
trees <- puremoe::data_mesh_trees()
```

```{r trees-table}
trees |>
  head(20) |>
  DT::datatable(rownames = FALSE, options = list(scrollX = TRUE))
```

## MeSH descriptor frequencies

`data_mesh_frequencies` is a bundled dataset giving the frequency of each MeSH
descriptor across the full PubMed corpus (39.7 M PMIDs, April 2026). Proportions
use the total corpus as denominator, making them suitable as a baseline for
enrichment analyses against arbitrary PubMed subsets.

```{r frequencies}
puremoe::data_mesh_frequencies |>
  head(20) |>
  DT::datatable(rownames = FALSE, options = list(scrollX = TRUE))
```

## Persistent storage

Both datasets are ~10 MB and fetched from GitHub on each call by default. To avoid re-downloading every session, set `use_persistent_storage = TRUE` — the files are cached to a system data directory and reused on subsequent calls.

```{r persistent, eval=FALSE}
thesaurus <- puremoe::data_mesh_thesaurus(use_persistent_storage = TRUE)
trees     <- puremoe::data_mesh_trees(use_persistent_storage = TRUE)
```

---
title: "Getting Started"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  message = FALSE,
  warning = FALSE,
  comment = "#>"
)
```

`puremoe` provides a unified interface to PubMed and NLM data. Search with `search_pubmed()`, then retrieve data from any of five endpoints with `get_records()`.

```{r libs}
library(puremoe)
library(dplyr)
library(DT)
```

## Search

`search_pubmed()` accepts standard PubMed query syntax and returns a vector of PMIDs.

```{r search}
pmids <- puremoe::search_pubmed('("political ideology"[TiAb])')
length(pmids)
```

```{r subset}
pmids_sub <- head(pmids, 50L)
```

## Abstracts

```{r abstracts}
abstracts <- puremoe::get_records(
  pmids_sub,
  endpoint = "pubmed_abstracts",
  cores    = 1L,
  sleep    = 0.5
)

abstracts <- abstracts |> mutate(pmid = as.character(pmid))
```

```{r abstracts-table}
abstracts |>
  select(pmid, year, journal, articletitle) |>
  DT::datatable(rownames = FALSE)
```

The `annotations` column is a list of per-article data frames containing MeSH terms, chemical names, and keywords.

```{r annotations}
bind_rows(abstracts$annotations) |>
  head(20) |>
  DT::datatable(rownames = FALSE)
```

## Affiliations

```{r affiliations}
affiliations <- puremoe::get_records(
  head(pmids_sub, 25L),
  endpoint = "pubmed_affiliations",
  cores    = 1L,
  sleep    = 0.5
)

affiliations |>
  DT::datatable(rownames = FALSE)
```

## iCite metrics

```{r icites}
icites <- puremoe::get_records(
  pmids_sub,
  endpoint = "icites",
  cores    = 1L,
  sleep    = 0.25
)

icites |>
  mutate(pmid = as.character(pmid)) |>
  select(-citation_net, -cited_by_clin) |>
  DT::datatable(rownames = FALSE, options = list(scrollX = TRUE))
```

## PubTator annotations

```{r pubtations}
pubtations <- puremoe::get_records(
  head(pmids_sub, 30L),
  endpoint = "pubtations",
  cores    = 1L
)

pubtations |>
  DT::datatable(rownames = FALSE)
```

## Full text

Full-text retrieval requires open-access PMC articles. `pmid_to_ftp()` resolves
PMIDs to XML URLs via the PMC Cloud Service on AWS S3, filtering to only those
with open-access full text available. In August 2026, NCBI will complete its
migration from the legacy PMC FTP Service to the Cloud Service; `puremoe` already
uses the new service.

```{r ftp}
ftp <- puremoe::pmid_to_ftp(pmids = pmids_sub)
ftp |> DT::datatable(rownames = FALSE, options = list(scrollX = TRUE))
```

```{r fulltext}
fulltext <- puremoe::get_records(
  head(ftp$url, 2L),
  endpoint = "pmc_fulltext",
  cores    = 1L
)

fulltext |>
  mutate(text = sapply(strsplit(text, "\\s+"), function(w) paste0(paste(head(w, 15), collapse = " "), "..."))) |>
  slice(1:5) |>
  DT::datatable(rownames = FALSE, options = list(scrollX = TRUE))
```

## Endpoint schemas

`endpoint_info()` returns column definitions, rate limits, and notes for any endpoint.

```{r endpoint-info}
puremoe::endpoint_info()
```

```{r endpoint-detail}
puremoe::endpoint_info("icites")
```

---
title: "Test: PMC Full-Text Retrieval"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE, message = TRUE, warning = TRUE)
library(puremoe)
```

## Step 0 -- 

```{r}
pmids <- puremoe::search_pubmed(x = "'political ideology'[TiAb]")[1:50]
```


## Step 1 — PMIDs to FTP URLs

```{r pmid_to_ftp}
## pmids <- c("11250746", "11573492", "12345678", "33515445")

ftp <- puremoe::pmid_to_ftp(pmids, verbose = TRUE)
ftp
```

## Step 2 — Retrieve Full Text

```{r source_pmc}
ft <- puremoe::get_records(ftp$url, 
                           endpoint = "pmc_fulltext", 
                           cores = 1)

ft
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search_pubmed.R
\name{search_pubmed}
\alias{search_pubmed}
\title{Search 'PubMed' Records}
\usage{
search_pubmed(
  x,
  start_year = NULL,
  end_year = NULL,
  retmax = 9999,
  use_pub_years = FALSE
)
}
\arguments{
\item{x}{Character string, the search query.}

\item{start_year}{Integer, the start year of publication date range (used if `use_pub_years` is TRUE).}

\item{end_year}{Integer, the end year of publication date range (used if `use_pub_years` is TRUE).}

\item{retmax}{Integer, maximum number of records to retrieve, defaults to 9999.}

\item{use_pub_years}{Logical, whether to filter search by publication years, defaults to TRUE.}
}
\value{
Numeric vector of unique PubMed IDs.
}
\description{
Performs a 'PubMed' search based on a query, optionally filtered by publication years. 
Returns a unique set of 'PubMed' IDs matching the query.
}
\examples{
ethnob1 <- search_pubmed("ethnobotany", 2010, 2012)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pmid_to_ftp.R
\name{pmid_to_ftp}
\alias{pmid_to_ftp}
\title{Convert PubMed IDs (PMIDs) to PMC IDs and Full-Text URLs}
\usage{
pmid_to_ftp(
  pmids,
  batch_size = 200L,
  sleep = 0.5,
  verbose = FALSE,
  ncbi_key = NULL
)
}
\arguments{
\item{pmids}{A character or numeric vector of PubMed IDs (PMIDs) to convert.}

\item{batch_size}{An integer specifying the number of PMIDs to process per
batch for ID conversion. Defaults to 200L. The NCBI API has limitations on batch sizes.}

\item{sleep}{A numeric value specifying the number of seconds to pause
between API requests for ID conversion (Step 1). Defaults to 0.5 seconds.
For OA API calls (Step 2), sleep time is automatically adjusted based on
rate limits: 0.11s with API key (10 req/sec), 0.34s without (3 req/sec).}

\item{verbose}{Logical, whether to print progress messages. Defaults to FALSE.}

\item{ncbi_key}{(Optional) NCBI API key for authenticated access.}
}
\value{
A data.table with columns:
  \itemize{
    \item \code{pmid}: The input PubMed ID
    \item \code{pmcid}: The corresponding PMC ID
    \item \code{doi}: The corresponding DOI (NA if not available)
    \item \code{url}: The full HTTPS URL for downloading PMC full text
  }
  Results are filtered to only include rows with valid URLs (open access articles),
  ordered by PMID. Returns NULL with a message if the API is unavailable or returns invalid data.
}
\description{
This function converts PMIDs to PMC IDs, then fetches the full-text file URLs
from the PMC Open Access service. It combines both steps into a single workflow.
}
\examples{
\donttest{
if (interactive()) {
  # Convert PMIDs to PMC IDs and get full-text URLs
  result <- pmid_to_ftp(c("11250746", "11573492"))
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_mesh_thesaurus.R
\name{data_mesh_thesaurus}
\alias{data_mesh_thesaurus}
\title{Download and Combine 'MeSH' and Supplemental Thesauruses}
\usage{
data_mesh_thesaurus(
  path = NULL,
  use_persistent_storage = FALSE,
  force_install = FALSE
)
}
\arguments{
\item{path}{A character string specifying the directory path where data should 
be stored. If not provided and persistent storage is requested, it defaults to 
a system-appropriate persistent location managed by `rappdirs`.}

\item{use_persistent_storage}{A logical value indicating whether to use persistent
storage. If TRUE and no path is provided, data will be stored in a system-appropriate 
location. Defaults to FALSE, using a temporary directory.}

\item{force_install}{A logical value indicating whether to force re-downloading 
of the data even if it already exists locally.}
}
\value{
A data.table containing the combined MeSH and supplemental thesaurus data.
}
\description{
This function downloads and combines the 'MeSH' (Medical Subject Headings) Thesaurus 
and a supplemental concept thesaurus.
The data is sourced from specified URLs and stored locally for subsequent use.
By default, the data is stored in a temporary directory. Users can opt into 
persistent storage by setting `use_persistent_storage` to TRUE and optionally 
specifying a path.
}
\examples{
\donttest{
if (interactive()) {
  data <- data_mesh_thesaurus()
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_records.R
\name{get_records}
\alias{get_records}
\title{Retrieve Data from 'NLM'/'PubMed' databases Based on PMIDs}
\usage{
get_records(
  pmids,
  endpoint = c("pubtations", "icites", "pubmed_affiliations", "pubmed_abstracts",
    "pmc_fulltext"),
  cores = 3,
  sleep = 1,
  ncbi_key = NULL
)
}
\arguments{
\item{pmids}{A vector of PMIDs for which data is to be retrieved. For 'pmc_fulltext' endpoint, 
provide full URLs instead (e.g., from \code{pmid_to_pmc()$url}).}

\item{endpoint}{A character vector specifying the type of data to retrieve ('pubtations', 'icites',
'pubmed_affiliations', 'pubmed_abstracts', 'pmc_fulltext').}

\item{cores}{Number of cores to use for parallel processing (default is 3).}

\item{sleep}{Duration (in seconds) to pause after each batch}

\item{ncbi_key}{(Optional) NCBI API key for authenticated access.}
}
\value{
A data.table containing combined results from the specified endpoint.
}
\description{
This function retrieves different types of data (like 'PubMed' records, affiliations, 'iCites 'data, etc.) from 'PubMed' based on provided PMIDs. It supports parallel processing for efficiency.
}
\details{
For the 'pmc_fulltext' endpoint, provide full URLs to PMC tar.gz files. 
Use \code{\link{pmid_to_pmc}} to convert PMIDs to PMC IDs and full URLs first.
}
\examples{
pmids <- c("38136652")
results <- get_records(pmids, endpoint = "pubmed_abstracts", cores = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_mesh_frequencies.R
\docType{data}
\name{data_mesh_frequencies}
\alias{data_mesh_frequencies}
\title{MeSH Descriptor Frequencies Across PubMed}
\format{
A data.table with 30,521 rows and 4 columns:
\describe{
  \item{DescriptorUI}{MeSH descriptor unique identifier (e.g.,
    \code{D000001})}
  \item{DescriptorName}{Canonical MeSH descriptor name}
  \item{n_pmids}{Number of distinct PubMed records indexed with this
    descriptor}
  \item{prop_total}{Proportion of all 39,703,112 PubMed PMIDs indexed
    with this descriptor}
}
}
\source{
Computed from \code{mesh_descriptor} table in a local PubMed
  PostgreSQL mirror; descriptor metadata from the NLM MeSH Thesaurus
  (April 2026).
}
\usage{
data_mesh_frequencies
}
\description{
Baseline frequencies for MeSH descriptors computed from a local PostgreSQL
mirror of PubMed (April 2026). For each descriptor, counts reflect the
number of distinct PMIDs indexed with that term; proportions use the full
PubMed corpus of 39,703,112 PMIDs as denominator. Descriptor UI and
canonical name are joined from the NLM MeSH thesaurus.
Intended as a baseline for MeSH term enrichment analyses against arbitrary
PubMed subsets.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_pmc_list.R
\name{data_pmc_list}
\alias{data_pmc_list}
\title{Download and Load 'PMC' Open Access File List}
\usage{
data_pmc_list(
  path = NULL,
  use_persistent_storage = FALSE,
  force_install = FALSE
)
}
\arguments{
\item{path}{A character string specifying the directory path where data should 
be stored. If not provided and persistent storage is requested, it defaults to 
a system-appropriate persistent location managed by `rappdirs`.}

\item{use_persistent_storage}{A logical value indicating whether to use persistent
storage. If TRUE and no path is provided, data will be stored in a system-appropriate 
location. Defaults to FALSE, using a temporary directory.}

\item{force_install}{A logical value indicating whether to force re-downloading 
of the data even if it already exists locally.}
}
\value{
A data.table containing the PMC file list with columns: file_path, 
citation, pmcid, pmid, and license_code.
}
\description{
This function downloads and loads the 'PMC' (PubMed Central) Open Access file list.
The file list contains mappings between PMC IDs, PMIDs, and file paths for 
open access articles available for download.
}
\details{
The data is sourced from NCBI's FTP server and stored locally for subsequent use.
By default, the data is stored in a temporary directory. Users can opt into 
persistent storage by setting `use_persistent_storage` to TRUE and optionally 
specifying a path.
}
\examples{
\donttest{
if (interactive()) {
  data <- data_pmc_list()
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pmid_to_pmc.R
\name{pmid_to_pmc}
\alias{pmid_to_pmc}
\title{Convert PubMed IDs (PMIDs) to PMC IDs}
\usage{
pmid_to_pmc(pmids, batch_size = 200L, sleep = 0.5)
}
\arguments{
\item{pmids}{A character or numeric vector of PubMed IDs (PMIDs) to convert.}

\item{batch_size}{An integer specifying the number of PMIDs to process per
batch. Defaults to 200L. The NCBI API has limitations on batch sizes.}

\item{sleep}{A numeric value specifying the number of seconds to pause
between API requests. Defaults to 0.5 seconds to respect API rate limits.}
}
\value{
A data.table with columns:
  \itemize{
    \item \code{pmid}: The input PubMed ID
    \item \code{pmcid}: The corresponding PMC ID (NA if not available in PMC)
    \item \code{doi}: The corresponding DOI (NA if not available)
  }
  Results are ordered by PMID. Returns NULL with a message if the API is
  unavailable or returns invalid data.
}
\description{
This function converts a vector of PubMed IDs (PMIDs) to their corresponding
PubMed Central (PMC) IDs and DOIs using the NCBI ID Converter API.
}
\examples{
\donttest{
if (interactive()) {
  # Convert a single PMID to PMC ID
  result <- pmid_to_pmc("12345678")
  
  # Convert multiple PMIDs
  pmids <- c("12345678", "23456789", "34567890")
  result <- pmid_to_pmc(pmids, batch_size = 10, sleep = 1)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoint_info.R
\name{endpoint_info}
\alias{endpoint_info}
\title{Get Information About Available Endpoints}
\usage{
endpoint_info(endpoint = NULL, format = c("list", "json"))
}
\arguments{
\item{endpoint}{Character string specifying which endpoint to get information about.
If NULL (default), returns a list of all available endpoints.}

\item{format}{Character string specifying the output format. Either "list" (default)
or "json" for JSON-formatted output.}
}
\value{
If \code{endpoint} is NULL, returns a character vector of available endpoint names.
  If \code{endpoint} is specified, returns a list (or JSON string) with detailed information
  about that endpoint including description, columns, parameters, rate limits, and notes.
}
\description{
This function provides detailed information about the available endpoints
in the package, including column descriptions, parameters, rate limits, and usage notes.
}
\examples{
\donttest{
if (interactive()) {
  # List all available endpoints
  endpoint_info()
  
  # Get information about a specific endpoint
  endpoint_info("pubmed_abstracts")
  
  # Get information in JSON format
  endpoint_info("icites", format = "json")
}
}

}
